@@ -1,8 +1,9 @@
 package HTML::Template;
 
 use strict;
+use integer;
 use vars qw( $VERSION %CACHE );
-$VERSION = 0.06;
+$VERSION = 0.9;
 
 %CACHE = ();
 
@@ -53,13 +54,14 @@ My Path is set to /bin;/usr/bin
 =head1 DESCRIPTION
 
 This module attempts make using HTML templates simple and natural.  It
-extends standard HTML with a few new HTML-esque tags - <TMPL_VAR> and
-<TMPL_LOOP>.  The file written with HTML and these new tags is called
-a template.  It is usually saved separate from your script - possibly
-even created by someone else!  Using this module you fill in the
-values for the variables and loops declared in the template.  This
-allows you to seperate design - the HTML - from the data, which you
-generate in the Perl script.
+extends standard HTML with a few new HTML-esque tags - <TMPL_VAR>,
+<TMPL_LOOP>, <TMPL_INCLUDE>, <TMPL_IF> and <TMPL_ELSE>.  The file
+written with HTML and these new tags is called a template.  It is
+usually saved separate from your script - possibly even created by
+someone else!  Using this module you fill in the values for the
+variables, loops and branches declared in the template.  This allows
+you to seperate design - the HTML - from the data, which you generate
+in the Perl script.
 
 This module is licenced under the GPL.  See the LICENSE section
 below for more details.
@@ -107,6 +109,9 @@ like:
 is not really valid HTML, but it is a perfectly valid use and will
 work as planned.
 
+The "NAME=" in the tag is optional, although for extensibility's sake I
+recommend using it.  Example - "<TMPL_LOOP LOOP_NAME>" is acceptable.
+
 =head2 <TMPL_VAR NAME="PARAMETER_NAME">
 
 The <TMPL_VAR> tag is very simple.  For each <TMPL_VAR> tag in the
@@ -115,10 +120,6 @@ the template is output the <TMPL_VAR> is replaced with the VALUE text
 you specified.  If you don't set a parameter it just gets skipped in
 the output.
 
-The "NAME=" in the tag is optional, although for extensibility's sake I
-recommend using it.  Example - "<TMPL_VAR PARAMETER_NAME>" is
-acceptable.
-
 =head2 <TMPL_LOOP NAME="LOOP_NAME"> </TMPL_LOOP>
 
 The <TMPL_LOOP> tag is a bit more complicated.  The <TMPL_LOOP> tag
@@ -180,22 +181,64 @@ name=>value pairs for a single pass over the loop template.  It is
 probably in your best interest to build these up programatically, but
 that is up to you!
 
-The "NAME=" in the tag is optional, although for extensibility's sake I
-recommend using it.  Example - "<TMPL_LOOP LOOP_NAME>" is acceptable.
-
 =head2 <TMPL_INCLUDE NAME="filename.tmpl">
 
 This tag includes a template directly into the current template at the
-point where the tag is found.  The included template contents work
-exactly as if its contents were physically included in the master
-template.
-
-The "NAME=" in the tag is optional, although for extensibility's sake I
-recommend using it.  Example - "<TMPL_INCLUDE filename.tmpl>" is
-acceptable.
+point where the tag is found.  The included template contents exactly
+as if its contents were physically included in the master template.
 
 NOTE: Currently, each <TMPL_INCLUDE> must be on a seperate line by itself.
 
+=head2 <TMPL_IF NAME="CONTROL_PARAMETER_NAME"> </TMPL_IF>
+
+The <TMPL_IF> tag allows you to include or not include a block of the
+template based on the value of a given parameter name.  If the
+parameter is given a value that is true for Perl - like '1' - then the
+block is included in the output.  If it is not defined, or given a
+false value - like '0' - then it is skipped.  The parameters are
+specified the same way as with TMPL_VAR.
+
+Example Template:
+
+   <TMPL_IF NAME="BOOL">
+     Some text that only gets displayed if BOOL is true!
+   </TMPL_IF>
+
+Now if you call $template->param(BOOL => 1) then the above block will
+be included by output. 
+
+<TMPL_IF> </TMPL_IF> blocks can include any valid HTML::Template
+construct - VARs and LOOPs and other IF/ELSE blocks.  Note, however,
+that intersecting a <TMPL_IF> and a <TMPL_LOOP> is invalid.
+
+   Example !!!THIS EXAMPLE WILL NOT WORK!!! :
+
+   <TMPL_IF BOOL>
+      <TMPL_LOOP SOME_LOOP>
+   </TMPL_IF>
+      </TMPL_LOOP>
+
+WARNING: Much of the benefit of HTML::Template is in decoupling your
+Perl and HTML.  If you introduce numerous cases where you have
+TMPL_IFs and matching Perl if()s, you will create a maintainance
+problem in keeping the two synchronized.  I suggest you adopt the
+practice of only using TMPL_IF if you can do so without requiring a
+matching if() in your Perl code.
+
+=head2 <TMPL_ELSE>
+
+You can include an alternate block in your TMPL_IF block by using
+TMPL_ELSE.  NOTE: You still end the block with </TMPL_IF>, not
+</TMPL_ELSE>!
+ 
+   Example:
+
+   <TMPL_IF BOOL>
+     Some text that is included only if BOOL is true
+   <TMPL_ELSE>
+     Some text that is included only if BOOL is false
+   </TMPL_IF>
+
 =cut
 
 =head1 Methods
@@ -209,9 +252,9 @@ Call new() to create a new Template object:
                                     );
 
 You must call new() with at least one name => value pair specifing how
-
 to access the template text.  You can use "filename => 'file.tmpl'" to
-specify a filename to be opened as the template. Alternately you can use:
+specify a filename to be opened as the template.  Alternately you can
+use:
 
   my $t = HTML::Template->new( scalarref => $ref_to_template_text, 
                                option => 'value' 
@@ -245,8 +288,7 @@ prefer.
 And as a final option, for those that might prefer it, you can call new as:
 
   my $t = HTML::Template->new_file(type => 'filename', 
-                                   source => 'file.tmpl',
-                                   option => 'value');
+                                   source => 'file.tmpl');
 
 Which works for all three of the source types.
 
@@ -403,22 +445,43 @@ sub new {
     die "HTML::Template->new called with multiple (or no) template sources specified!  A valid call to new() has exactly one filename => 'file' OR exactly one scalarRef => \\\$scalar OR exactly one arrayRef = \\\@array";    
   }
 
-  # deal with $ENV{HTML_TEMPLATE_ROOT} stuff
-  if (defined($options->{filename})) {
-    if (!($options->{filename} =~ /^\//) and exists($ENV{HTML_TEMPLATE_ROOT})) {
-      $options->{filename} = $ENV{HTML_TEMPLATE_ROOT} . '/' . $options->{filename};
-    }
-  }
-
-  # Go!
-  $options->{timing} and $self->{timer}{start_time} = (times)[0];
-  
   # initialize data structures
   $self->_init;
   
   return $self;
 }
 
+# an internally used new that recieves its parse_stack and param_map as input
+sub _new_from_loop {
+  my $pkg = shift;
+  my $self; { my %hash; $self = bless(\%hash, $pkg); }
+
+  # the options hash
+  my $options = {};
+  $self->{options} = $options;
+
+  # set default parameters in options hash
+  $options->{debug} = 0;
+  $options->{cache} = 0;
+  $options->{blind_cache} = 0;
+  $options->{cache_debug} = 0;
+  $options->{die_on_bad_params} = 1;
+  $options->{vanguard_compatibility_mode} = 0;
+  $options->{no_includes} = 0;
+  $options->{associate} = [];
+
+  # load in options supplied to new()
+  for (my $x = 0; $x <= $#_; $x += 2) { 
+    defined($_[($x + 1)]) or die "HTML::Template->new() called with odd number of option parameters - should be of the form option => value";
+    $options->{lc($_[$x])} = $_[($x + 1)]; 
+  }
+
+  $self->{param_map} = $options->{param_map};
+  $self->{parse_stack} = $options->{parse_stack};
+
+  return $self;
+}
+
 # a few shortcuts to new(), of possible use...
 sub new_file {
   my $pkg = shift; return $pkg->new('filename', @_);
@@ -436,8 +499,12 @@ sub _init {
   my $self = shift;
   my $options = $self->{options};
 
-  # initialize the parameter values hash
-  $self->{param_values} = {};
+  # deal with $ENV{HTML_TEMPLATE_ROOT} stuff
+  if (defined($options->{filename})) {
+    if (!($options->{filename} =~ /^\//) and exists($ENV{HTML_TEMPLATE_ROOT})) {
+      $options->{filename} = $ENV{HTML_TEMPLATE_ROOT} . '/' . $options->{filename};
+    }
+  }
 
   # look in the cache to see if we have a cached copy of this
   # template, note modification time in $mtime if we do.
@@ -449,6 +516,7 @@ sub _init {
     
     if (!$options->{blind_cache}) {
       # make sure it still exists in the filesystem 
+
       (-r $options->{filename}) or die("HTML::Template : template file $options->{filename} does not exist or is unreadable.");    
 
       # get the modification time
@@ -487,32 +555,32 @@ sub _init {
     # else, use the cached values instead of calling _init_template
     # and _pre_parse.
     
-    $options->{cache_debug} and warn "CACHE HIT : $options->{filename} : $mtime";
+    $options->{cache_debug} and warn "CACHE HIT : $options->{filename}";
     
-    $self->{template} = $CACHE{$options->{filename}}{template};
     $self->{param_map} = $CACHE{$options->{filename}}{param_map};
-    $self->{loop_heap} = $CACHE{$options->{filename}}{loop_heap};
+    $self->{parse_stack} = $CACHE{$options->{filename}}{parse_stack};
     exists($CACHE{$options->{filename}}{included_mtimes}) and
       $self->{included_mtimes} = $CACHE{$options->{filename}}{included_mtimes};
+
+    # clear out values from param_map from last run
+    $self->clear_params();
+
     return $self;
   }
 
   # init the template and parse data
   $self->_init_template;
-  $options->{timing} and $self->{timer}{load_time} = (times)[0];
   $self->_pre_parse;
-  $options->{timing} and $self->{timer}{parse_time} = (times)[0];
 
   # if we're caching, cache the results of _init_template and _pre_parse
   # for future use
   if ($options->{cache} and exists($options->{filename})) {
-    $options->{cache_debug} and warn "CACHE LOAD : $options->{filename} : $mtime";
+    $options->{cache_debug} and warn "CACHE LOAD : $options->{filename}";
     
     $options->{blind_cache} or
       $CACHE{$options->{filename}}{mtime} = (stat($options->{filename}))[9];
-    $CACHE{$options->{filename}}{template} = $self->{template};
     $CACHE{$options->{filename}}{param_map} = $self->{param_map};
-    $CACHE{$options->{filename}}{loop_heap} = $self->{loop_heap};
+    $CACHE{$options->{filename}}{parse_stack} = $self->{parse_stack};
     exists($self->{included_mtimes}) and
       $CACHE{$options->{filename}}{included_mtimes} = $self->{included_mtimes};
   }
@@ -559,9 +627,11 @@ sub _init_template {
   }
 
   # look for TMPL_INCLUDEs and process them now
+  use vars qw($line);
+  local(*line);
   if (!($options->{no_includes})) {
     for (my $line_number = 0; $line_number <= $#{$self->{template}}; $line_number++) {
-      my $line = $self->{template}[$line_number];
+      *line = \$self->{template}[$line_number];
       defined($line) or next;
 
       if ($line =~ /<[tT][mM][pP][lL]_[Ii][Nn][Cc][Ll][Uu][Dd][Ee]\s+(?:[nN][aA][mM][eE]\s*=)?\s*"?([\w\/\.]+)"?\s*>/x) {
@@ -597,7 +667,7 @@ sub _init_template {
         my @lines = splice(@{$self->{template}}, ($line_number + 1), ($lines_in_existing_template - $line_number + 1));
         splice(@{$self->{template}}, ($line_number + 1), (scalar(@templateArray) + scalar(@lines) + 1), (@templateArray, @lines));
 
-        $self->{template}[$line_number] = '';
+        undef $self->{template}[$line_number];
       }
     }
   }
@@ -606,14 +676,7 @@ sub _init_template {
 }
 
 # _pre_parse sifts through a template building up the param_map and
-# loop_heap structures
-#
-# The param_map stores the names and location of TMPL_VAR parameters.
-# When output runs it can then just use the param_map to find
-# out where to make its substitutions.
-#
-# The loop_heap is a little more complicated.  It stores both the
-# location and the Template object for each TMPL_LOOP encountered.
+# parse_stack structures
 #
 # The end result is a Template object that is fully ready for
 # output().
@@ -622,127 +685,266 @@ sub _pre_parse {
   my $options = $self->{options};
   
   $options->{debug} and print "\nIn pre_parse:\n\n";
-
-  $options->{timing} and $self->{timer}{parse_in} = (times)[0];
-
-  $self->{param_map} = {};
-  $self->{loop_heap} = {};
   
-  for (my $line_number = 0; $line_number <= $#{$self->{template}}; $line_number++) {
+  # setup the stacks and maps - they're accessed by typeglobs that
+  # reference the end of the stack.
+  use vars qw(@pstack %pmap @ifstack);
+  local(*pstack, *ifstack, *pmap);
+  
+  my @pstacks = ([]);
+  *pstack = $pstacks[0];
+  $self->{parse_stack} = $pstacks[0];
+  
+  my @pmaps = ({});
+  *pmap = $pmaps[0];
+  $self->{param_map} = $pmaps[0];
+
+  my @ifstacks = ([]);
+  *ifstack = $ifstacks[0];
+
+  my @loopstack = ();
+
+  # loop through lines, filling up pstack
+ LINE: for (my $line_number = 0; $line_number <= $#{$self->{template}}; $line_number++) {
+    next unless defined $self->{template}[$line_number]; 
     my $line = $self->{template}[$line_number];
-    defined($line) or next;
-    my $done_with_line = 0;
 
     # handle the old vanguard format
-    if ($options->{vanguard_compatibility_mode}) {
-      if ($line =~ s/%([\w]+)%/<TMPL_VAR NAME=$1>/g) {
-        $self->{template}[$line_number] = $line;
-      }
-    }
-
-    while(!$done_with_line) {
-      # Look for a loop start
-      defined($line) or last;
+    $options->{vanguard_compatibility_mode} and 
+      $line =~ s/%([\w]+)%/<TMPL_VAR NAME=$1>/g;
 
+    # make PASSes over the line until nothing is left
+  PASS: while(1) {
+      last PASS unless defined($line);
+      
       # [tT]... is twice as fast as the 'i' regex option!
-      if ($line =~ /(.*?)<[tT][mM][pP][lL]_[lL][oO][oO][pP]\s+([nN][aA][mM][eE]\s*=)?\s*"?(\w+)"?\s*>(.*)/g) {
-        my $preloop = $1;
+      if ($line =~ /
+                    (.*?)
+                    (
+                      <\/?[tT][mM][pP][lL]_
+                      (?:
+                         (?:[Vv][Aa][Rr])
+                         |
+                         (?:[lL][oO][oO][pP])
+                         |
+                         (?:[Ii][Ff])
+                         |
+                         (?:[Ee][Ll][Ss][Ee])
+                      )
+                    )
+                    (?:
+                       \s+
+                       [nN][aA][mM][eE]
+                       \s*
+                       =
+                    )?
+                    \s*
+                    "?
+                    (
+                     \w+
+                    )?
+                    "?
+                    \s*
+                    >
+                    (.*)
+                   /gx) {
+        my $pre = $1;
+        my $which = $2;
         my $name = lc $3;
-        my $chunk = $4;
-        $options->{debug} and print "$line_number : saw loop $name\n";
-        $options->{timing} and $self->{timer}{loop_in} = (times)[0];      
-        
-        # find the end of the loop
-        my ($loop_body_ref, $leftover_ref, $pos);
-        for ($pos = $line_number; $pos <= $#{$self->{template}}; $pos++) {
-          if ($pos != $line_number) {
-            $chunk .= $self->{template}[$pos];
-          }
-          ($loop_body_ref, $leftover_ref) = $self->_extractLoop(\$chunk);
-          defined($loop_body_ref) and last;
-        }
-        (defined($loop_body_ref)) or die("HTML::Template : Problem looking for matching </TMPL_LOOP> for <TMPL_LOOP NAME=${name}> : Could not find one!");
-        $options->{debug} and print "Loop $name body: \n$$loop_body_ref\n\n";
-        
-        # store the results
-        push(@{$self->{loop_heap}{$name}{spot}}, $line_number);
-        push(@{$self->{loop_heap}{$name}{template_object}}, HTML::Template->new( scalarref => $loop_body_ref, debug => $options->{debug}, die_on_bad_params => $options->{die_on_bad_params}, no_includes => 1));
+        my $post = $4;
+
+        if ($which =~ /^<[tT][mM][pP][lL]_[Vv][Aa][Rr]/) {
+          $options->{debug} and print "$line_number : saw $name : \"$line\"\n";
           
-        # if we've got a multiline match we'll need to undef the
-        # lines we gobbled for the loop body.
-        if ($pos > $line_number) {
-          foreach (my $x = $line_number + 1; $x <= $pos; $x++) {
-            $self->{template}[$x] = undef;
+          # if we already have this var, then simply link to the existing
+          # HTML::Template::VAR, else create a new one.        
+          my $var;        
+          if (exists $pmap{$name}) {
+            $var = $pmap{$name};
+          } else {
+            $var = HTML::Template::VAR->new();
+            $pmap{$name} = $var;
+          }
+
+          # push results
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+          push(@pstack, $var);
+
+          # keep working on line
+          $line = $post;         
+          next PASS;          
+        
+        } elsif ($which =~ /^<[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp]/) {
+          # we've got a loop start
+          $options->{debug} and print "$line_number : saw loop $name\n";
+          
+          # if we already have this loop, then simply link to the existing
+          # HTML::Template::LOOP, else create a new one.
+          my $loop;
+          if (exists $pmap{$name}) {
+            $loop = $pmap{$name};
+          } else {
+            # store the results in a LOOP object - actually just a
+            # thin wrapper around another HTML::Template object.
+            $loop = HTML::Template::LOOP->new();
+            $pmap{$name} = $loop;
+          }
+
+          # get it on the loopstack
+          push(@pstack, $loop);
+          push(@loopstack, $loop);
+
+          # push results into parse_stack
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+
+          # magic time - push on fresh pmap and pstack, adjust the typeglobs
+          push(@pstacks, []);
+          *pstack = $pstacks[$#pstacks];
+          push(@pmaps, {});
+          *pmap = $pmaps[$#pmaps];
+          push(@ifstacks, []);
+          *ifstack = $ifstacks[$#ifstacks];
+
+          # keep working on the rest
+          $line = $post;
+          next PASS;
+
+        } elsif ($which =~ /^<\/[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp]/) {
+          $options->{debug} and print "$line_number : saw /loop : \"$line\"\n";
+          
+          my $loop = pop(@loopstack);
+          die "HTML::Template->new() : found </TMPL_IF|ELSE> with no matching <TMPL_IF|ELSE> at $line_number!" unless defined $loop;
+
+                                                     
+          # push text on the stack - must do before pstack manip!
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+
+          # get pmap and pstack for the loop, adjust the typeglobs
+          my $param_map = pop(@pmaps);
+          *pmap = $pmaps[$#pmaps];
+          my $parse_stack = pop(@pstacks);
+          *pstack = $pstacks[$#pstacks];
+          
+          scalar(@ifstack) and die "HTML::Template->new() : Dangleing <TMPL_IF> in loop ending at $line_number!";
+          pop(@ifstacks);
+          *ifstack = $ifstacks[$#ifstack];
+          
+          # instantiate the sub-Template:
+          $loop->[0] = HTML::Template->_new_from_loop(
+                                                      parse_stack => $parse_stack,
+                                                      param_map => $param_map,
+                                                      debug => $options->{debug}, 
+                                                      die_on_bad_params => $options->{die_on_bad_params}, 
+                                                     );
+          
+          # next line
+          $line = $post;
+          next PASS;
+          
+        } elsif ($which =~ /^<[Tt][Mm][Pp][Ll]_[Ii][Ff]/) {
+          defined($name) or die "HTML::Template->new() : found TMPL_IF with no name at $line_number!";
+          $options->{debug} and print "$line_number : saw if $name : \"$line\"\n";
+          # if we already have this var, then simply link to the existing
+          # HTML::Template::VAR, else create a new one.        
+          my $var;        
+          if (exists $pmap{$name}) {
+            $var = $pmap{$name};
+          } else {
+            $var = HTML::Template::VAR->new();
+            $pmap{$name} = $var;
+          }
+
+          # connect the var to an if
+          my $if = HTML::Template::IF->new($var);
+
+          # push results
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+          push(@pstack, $if);
+          push(@ifstack, $if);
+
+          $line = $post;
+          next PASS;
+
+        } elsif ($which =~ /^<\/[Tt][Mm][Pp][Ll]_[Ii][Ff]/x) {
+          $options->{debug} and print "$line_number : saw /if : \"$line\"\n";
+
+          my $if = pop(@ifstack);
+          die "HTML::Template->new() : found </TMPL_IF|ELSE> with no matching <TMPL_IF|ELSE> at $line_number!" unless defined $if;
+
+          # push text on the stack
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+
+          # connect the matching to this "address"
+          $if->[1] = ($#pstack + 1);
+
+          # next line
+          $line = $post;
+          next PASS;
+
+        } elsif ($which =~ /^<[Tt][Mm][Pp][Ll]_[Ee][Ll][Ss][Ee]/) {
+          $options->{debug} and print "$line_number : saw else : \"$line\"\n";
+
+          my $if = pop(@ifstack);
+          die "HTML::Template->new() : found <TMPL_ELSE> with no matching <TMPL_IF> at $line_number!" unless defined $if;
+          
+          my $else = HTML::Template::ELSE->new($if->[0]);
+
+          # push stuff onto the stack
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+          push(@pstack, $else);
+          push(@ifstack, $else);
+
+          # connect the matching to this "address"
+          $if->[1] = ($#pstack);
+
+          $line = $post;
+          next PASS;
+
+        } else {
+          # zuh!?
+          die "HTML::Template->new() : Unknown or unmatched TMPL construct on line $line_number";
+        }
+
+      } else {
+        # push the rest and get a new line
+        if (defined($line)) {
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $line;
+          } else {
+            push(@pstack, \$line);
           }
         }
-        
-        # now reform $line to remove loop body
-        $line = $preloop . ' <TMPL_LOOP NAME=' . $name . ' PLACEHOLDER> ' . $$leftover_ref;
-        # donate back the changes
-        $self->{template}[$line_number] = $line;
-
-        # tick
-        if ($options->{timing}) {
-          $self->{timer}{loop_out} = (times)[0];
-          warn "Loop $name : find time " . ($self->{timer}{loop_out} - $self->{timer}{loop_in});
-        }
-
-        # next pass over line
-        next;          
+        next LINE;
       }
-
-      defined($line) or last;
-      # [tT]... is twice as fast as the 'i' regex option!
-      my @names = ($line =~ /<[tT][mM][pP][lL]_[vV][aA][rR]\s+(?:[nN][aA][mM][eE]\s*=)?\s*"?(\w+)"?\s*>/gx);
-      foreach my $name (@names) {
-        $name = lc($name);
-        exists($self->{param_map}{$name}) or $self->{param_map}{$name} = [];
-        push(@{$self->{param_map}{$name}}, $line_number);
-        
-        # set their value initially to undef
-        $self->{param_values}{$name} = undef;
-        
-        $options->{debug} and print "$line_number : saw $name\n";
-      }
-
-      # all done
-      $done_with_line = 1;
     }
   }
 
-  return $self;
-}
-
-# returns (\$loop_body, \$leftover) given a chunk following directly
-# after a <TMPL_LOOP NAME=BLAH> tag.  Returns undef if the block does
-# not contain a valid loop body.
-sub _extractLoop {
-  my $self = shift;
-
-  use vars qw($chunk);
-  local *chunk = shift;
-
-  # a huge speedup on large templates - the loop body extraction regex
-  # is really quite slow.
-  # [tT]... is twice as fast as the 'i' regex option!
-  if (!( $chunk =~ /<\/[Tt][Mm][Pp][Ll]_[Ll][Oo][Oo][Pp]>/s)) {
-    return undef;
-  }
-
-  # [tT]... is twice as fast as the 'i' regex option!
-  while ($chunk =~ /(.+)<\/[Tt][Mm][Pp][Ll]_[Ll][Oo][Oo][Pp]>(.*)/gs) {
-    my $loop_body = $1;
-    my $leftover = $2;
-
-    # if the loop body has an equal number of loop starts and stops
-    # then it's a valid loop body!
-    if (scalar($loop_body =~ /(<[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp])/g) ==
-        scalar($loop_body =~ /(<\/[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp])/g)) {
-      # found it!
-      return (\$loop_body, \$leftover);
-    }
-  }
-  return undef;
+  # all done with template
+  delete $self->{template};
 }
 
 =head2 param
@@ -776,44 +978,54 @@ param() can be called in three ways
 
 =cut
 
+
 sub param {
   my ($self, $param, $value) = @_;
   my $options = $self->{options};
+  my $param_map = $self->{param_map};
 
   if (!defined($param)) {
     # return a list of parameters in this template
-    return (keys(%{$self->{param_map}}));
+    return keys(%{$param_map});
+
   } elsif (!defined($value)) {
     $param = lc($param);
+    
     # check for parameter existence 
-    if ($options->{die_on_bad_params} and !exists($self->{param_map}{$param})
-                                    and !exists($self->{loop_heap}{$param})) {
+    $options->{die_on_bad_params} and !exists($param_map->{$param}) and
       die("HTML::Template : Attempt to set nonexistent parameter $param");
-    }
+    
 
-    # return the value set to a param
-    return($self->{param_values}{$param});
+    return undef unless (exists($param_map->{$param}) and
+                         defined($param_map->{$param}));
+
+    (ref($param_map->{$param}) eq 'HTML::Template::VAR') and
+      return ${$param_map->{$param}};
+    (ref($param_map->{$param}) eq 'HTML::Template::LOOP') and
+      return $param_map->{$param}[1];
+    die "Unknown param type!";
   } else {
     $param = lc($param);
     # check that this param exists in the template
-    if ($options->{die_on_bad_params} and !exists($self->{param_map}{$param})
-                                    and !exists($self->{loop_heap}{$param})) {
+    $options->{die_on_bad_params} and !exists($param_map->{$param}) and
       die("HTML::Template : Attempt to set nonexistent parameter $param");
-    }
-
-    # set the parameter and return $self
-
+    
+    return unless (exists($param_map->{$param}) and
+                   defined($param_map->{$param}));
+    
     # copy in contents of ARRAY refs to prevent confusion - 
     # thanks Richard!
     if ( ref($value) eq 'ARRAY' ) {
-      $self->{param_values}{$param} = [@{$value}];
+      (ref($param_map->{$param}) eq 'HTML::Template::LOOP') or
+        die "HTML::Template::param() : attempt to set parameter $param with an array ref - parameter is not a TMPL_LOOP!";
+      $param_map->{$param}[1] = [@{$value}];
     } else {
-      $self->{param_values}{$param} = $value;
+      (ref($param_map->{$param}) eq 'HTML::Template::VAR') or
+        die "HTML::Template::param() : attempt to set parameter $param with a scalar - parameter is not a TMPL_VAR!";
+      ${$param_map->{$param}} = $value;
     }
     return $self;
   }
-
-  die("This can't be happening.");
 }
 
 =head2 clear_params()
@@ -824,20 +1036,22 @@ Sets all the parameters to undef.  Useful internally, if nowhere else!
 
 sub clear_params {
   my $self = shift;
-  foreach my $name ($self->param()) {
-    $self->{param_values}{$name} = undef;
+  foreach my $name (keys %{$self->{param_map}}) {
+    if (ref($self->{param_map}{$name}) eq 'HTML::Template::VAR') {      
+      undef ${$self->{param_map}{$name}};
+    } elsif (ref($self->{param_map}{$name}) eq 'HTML::Template::LOOP') {
+      undef $self->{param_map}{$name}[1];
+    }
   }
 }
 
 
-# obsolete implementation of associate functionality - still supported
+# obsolete implementation of associate
 sub associateCGI { 
   my $self = shift;
   my $cgi  = shift;
-  
   (ref($cgi) eq 'CGI') or
     die("Warning! non-CGI object was passed to HTML::Template::associateCGI()!\n");
-
   push(@{$self->{options}{associate}}, $cgi);
   return 1;
 }
@@ -864,104 +1078,129 @@ sub output {
   my $self = shift;
   my $options = $self->{options};
 
-  $options->{timing} and $self->{timer}{output_in_time} = (times)[0];
-  $options->{debug} and print "\nIn output\n\n";
+  $options->{debug} and print "\nIn output $self\n\n";
 
-  # keep a hash of lines changed in the replace loop
-  my %templateChanges;
-
-  # pre-extend to worst case - all lines are keys
-  keys(%templateChanges) = $#{$self->{template}};
-
-  # kick off the search and replace loop
-  # works by following the param_map for each named param
-  use vars qw($value);
-  local(*value);  
-  foreach my $name (keys %{$self->{param_map}}) {
-    *value = \$self->{param_values}{$name};
-
-    # support the associate magic.
-    if ( ! defined($value) ) {
-      foreach my $associated_object (@{$options->{associate}}) {
-        *value = \$associated_object->param($name);
-        defined($value) and last;
-      }
-    }
-
-    ($options->{debug} and !defined($value) 
-     and print "parameter $name not set at output()\n");
-    # visit each spot on the map and do a replace into templateChanges
-    foreach my $spot (@{$self->{param_map}{$name}}) {
-      defined($templateChanges{$spot}) 
-        or ($templateChanges{$spot} = $self->{template}[$spot]);
-      if (defined($value)) {
-        $templateChanges{$spot} =~ s/<tmpl_var\s+(?:name\s*=)?\s*"?${name}"?\s*>/$value/sgi;
-      } else {
-        $templateChanges{$spot} =~ s/<tmpl_var\s+(?:name\s*=)?\s*"?${name}"?\s*>//sgi;
-      }
-    }
-  }
-
-
-  # handle the loops
-  foreach my $name (keys %{$self->{loop_heap}}) {
-    my $valueARef = $self->{param_values}{$name};
-    ($options->{debug} and !defined($valueARef)) 
-      and print "parameter $name not set at output()\n";
-    
-    # visit each spot on the map and do a looping output() on the loop
-    # object.  Insert the result into the spot just like a var
-    # interpolation
-    for( my $x = 0; $x <= $#{$self->{loop_heap}{$name}{spot}}; $x++) {
-      my $spot = $self->{loop_heap}{$name}{spot}[$x];
-      my $tobj = $self->{loop_heap}{$name}{template_object}[$x];
-      defined($templateChanges{$spot}) 
-        or ($templateChanges{$spot} = $self->{template}[$spot]);
-      my $loop_output = '';
-      foreach my $valueSetRef (@{$valueARef}) {
-        # set the parameters for this iteration
-        foreach my $name (keys %{$valueSetRef}) {
-          $tobj->param($name, $valueSetRef->{$name});
+  # support the associate magic, searching for undefined params and
+  # attempting to fill them from the associated objects.
+  if (scalar(@{$options->{associate}})) {
+    foreach my $param (keys %{$self->{param_map}}) {
+      if (!defined($self->{param_map}->{$param}->param())) {
+      OBJ: foreach my $associated_object (@{$options->{associate}}) {
+          my $value = $associated_object->param($param);          
+          if (defined($value)) {
+            $self->{param_map}->{$param}->param($value);
+            last OBJ;
+          }
         }
-        # accumulate output and clear params
-        $loop_output .= $tobj->output;
-        $tobj->clear_params();
-      }
-      if (length($loop_output)) {
-        $templateChanges{$spot} =~ s/<tmpl_loop\s+name\s*=\s*"?${name}"?\s*PLACEHOLDER>/$loop_output/i;
-      } else {
-        $templateChanges{$spot} =~ s/<tmpl_loop\s+name\s*=\s*"?${name}"?\s*PLACEHOLDER>//i;
       }
     }
   }
 
-  # all done - concat up the resulting arrays, skipping undef'd lines
-  my $result = "";
-  for (my $x = 0; $x <= $#{$self->{template}}; $x++) {
-    if (exists($templateChanges{$x})) {
-      $result .= $templateChanges{$x};
-    } elsif (defined($self->{template}[$x])) {
-      $result .= $self->{template}[$x];
-    }
-  }
+  use vars qw($line @parse_stack); local(*line, *parse_stack);
 
-  # warn a little timing information if timing => 1
-  if ($options->{timing}) {
-    $self->{output_out_time} = (times)[0];
-    warn "Loaded Template at: " . ($self->{timer}{load_time} - $self->{timer}{start_time});
-    warn "Parsed Template at: " . ($self->{timer}{parse_time} - $self->{timer}{start_time});
-    warn "Params Loaded   at: " . ($self->{timer}{output_in_time} - $self->{timer}{start_time});
-    warn "Output Done     at: " . ($self->{timer}{output_out_time} - $self->{timer}{start_time});
+  # walk the parse stack, accumulating output in $result
+  *parse_stack = $self->{parse_stack};
+  my $result = '';
+  my $type;
+  for (my $x = 0; $x <= $#parse_stack; $x++) {
+    *line = \$parse_stack[$x];
+    $type = ref($line);
+
+    if ($type eq 'SCALAR') {
+      $result .= $$line;
+    } elsif ($type eq 'HTML::Template::VAR') {
+      defined($$line) and $result .= $$line;
+    } elsif ($type eq 'HTML::Template::LOOP') {
+      defined($line->[1]) and $result .= $line->output();
+    } elsif ($type eq 'HTML::Template::IF') {
+      (!defined(${$line->[0]}) or !${$line->[0]}) and
+        $x = $line->[1];
+    } elsif ($type eq 'HTML::Template::ELSE') {
+      (defined(${$line->[0]}) and ${$line->[0]}) and $x = $line->[1];
+    } else {
+      die "HTML::Template::output() : Unknown item in parse_stack : " . $type;
+    }
   }
 
   return $result;
 }
 
+# HTML::Template::VAR and LOOP are *light* objects - their internal
+# spec is used above.  No encapsulation or information hiding is to be
+# assumed.
+
+package HTML::Template::VAR;
+
+sub new {
+  my ($pkg) = @_;
+  my $value;
+  my $self = \$value;
+  bless($self, $pkg);
+  return $self;
+}
+
+sub param {
+  my ($self, $param) = @_;
+  defined($param) and $$self = $param;
+  return $$self;
+}
+
+package HTML::Template::LOOP;
+
+sub new {
+  my ($pkg) = shift;
+  my $self = [];
+  bless($self, $pkg);  
+  return $self;
+}
+
+sub param {
+  my ($self, $param) = @_;
+  defined($param) and $self->[1] = $param;
+  return $self->[1];  
+}
+
+sub output {
+  my $self = shift;
+  my $template = $self->[0];
+  my $value_sets_array = $self->[1];
+  next unless defined($value_sets_array);  
   
-  
+  my $result = '';
+  foreach my $value_set (@$value_sets_array) {
+    foreach my $name (keys %$value_set) {
+      $template->param($name, $value_set->{$name});
+    }
+    $result .= $template->output;
+    $template->clear_params;
+  }
+  return $result;
+}
+
+package HTML::Template::IF;
+
+sub new {
+  my $pkg = shift;
+  my $var = shift;
+  my $self = [$var];
+  bless($self, $pkg);  
+  return $self;
+}
+
+package HTML::Template::ELSE;
+
+sub new {
+  my $pkg = shift;
+  my $var = shift;
+  my $self = [$var];
+  bless($self, $pkg);  
+  return $self;
+}
+
 
 
 1;
+
 __END__
 
 =head1 BUGS
